#!/usr/bin/env python3
# encoding: utf-8
from tests.utils import *

def test_negative(usual_setup_usual_teardown):
    create_file('xxx', 'b.png')
    create_file('xxx', 'a.png')
    create_file('xxx', 'a')
    head, *data, footer = run_rmlint('-i')
    assert footer['total_files'] == 3
    assert footer['total_lint_size'] == 0
    assert footer['duplicates'] == 0


def test_positive(usual_setup_usual_teardown):
    create_file('xxx', 'a.png')
    create_file('xxx', 'a.jpg')
    head, *data, footer = run_rmlint('-i')
    assert footer['total_files'] == 2
    assert footer['total_lint_size'] == 3
    assert footer['duplicates'] == 1



import pytest
import tests.utils as utils

@pytest.fixture(autouse=True)
def with_cleanup_between_runs():
    utils.cleanup_testdir()
    utils.create_testdir()


@pytest.fixture(params=["sh", "bash", "dash"])
def shell(request):
    yield request.param


@pytest.fixture
def usual_setup_usual_teardown():
    utils.usual_setup_func()
    yield
    utils.usual_teardown_func()


@pytest.fixture
def usual_setup_mount_bind_teardown():
    utils.usual_setup_func()
    yield
    utils.mount_bind_teardown_func()



#!/usr/bin/env python3
import os
import subprocess

from tests.utils import *

FILE_SIZE_KB = 10000
DIFFERENT_BYTES = 1
KBYTES_FROM_END = 10

LARGE_FILE_SIZE = 5 * 1024**3  # 5 GiB


def test_bigfiles(usual_setup_usual_teardown):

    num_bytes = int(FILE_SIZE_KB * 1024)
    # create two identical files and a third one which differs near the end:
    create_file('x' * int(FILE_SIZE_KB * 1024), 'file1')

    create_file('x' * int(FILE_SIZE_KB * 1024), 'file2')

    create_file('x' * int((FILE_SIZE_KB - KBYTES_FROM_END )* 1024) +
                'y' * int(DIFFERENT_BYTES) +
                'x' * int(KBYTES_FROM_END * 1024 - DIFFERENT_BYTES),
                'file3')

    *_, footer = run_rmlint('')
    assert footer['duplicates'] == 1


def _setup_large_file_offset():
    if not has_feature('bigfiles'):
        pytest.skip('rmlint built without large file support')

    path_a = create_file('', 'a')
    path_b = create_file('', 'b')
    path_c = create_file('', 'c')

    os.truncate(path_a, 4 * 1024)
    if os.stat(path_a).st_blocks:
        # only really works on Linux
        pytest.skip('cannot make sparse files with truncate()')

    # allocate large sparse files
    os.truncate(path_a, LARGE_FILE_SIZE)
    os.truncate(path_b, LARGE_FILE_SIZE)
    os.truncate(path_c, LARGE_FILE_SIZE)

    # touch last byte of one file
    with open(path_a, 'r+') as f:
        f.seek(LARGE_FILE_SIZE - 1)
        f.write('x')

    return path_a, path_b, path_c


def test_hash_utility(usual_setup_usual_teardown):
    path_a, path_b, path_c = _setup_large_file_offset()

    # only files 'b' and 'c' should match
    # metro is chosen because it's faster
    output = subprocess.check_output([
        *'./rmlint --hash -a metro'.split(),
        path_a, path_b, path_c,
    ])
    hashes = [l.split()[0] for l in output.splitlines()]
    assert hashes[0] != hashes[1]  # a != b
    assert hashes[1] == hashes[2]  # b == c



#!/usr/bin/env python3
import os

from tests.utils import *

SOURCE = '''
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>

int main(int argc, char *argv[]) {
    puts("Hello rmlint. Why were you executing this?");
    return EXIT_SUCCESS;
}
'''


def create_binary(path, stripped=False):
    path = path + '.stripped' if stripped else path + '.nonstripped'
    full_path = os.path.join(TESTDIR_NAME, path)

    command = '{cc} -o {path} {option} -std=c99 -xc -'.format(
        cc=os.environ.get('CC', 'gcc'), path=full_path, option='-s' if stripped else '-ggdb3',
    )
    subprocess.run(command, input=SOURCE, shell=True, universal_newlines=True, check=True)


def test_negative(usual_setup_usual_teardown):
    if has_feature('nonstripped') is False:
        return

    create_file(SOURCE, 'source.c')
    create_binary('source.c', stripped=True)
    head, *data, footer = run_rmlint('-T "none +nonstripped"')
    assert footer['total_files'] == 2
    assert footer['total_lint_size'] == 0
    assert len(data) == 0


def test_positive(usual_setup_usual_teardown):
    if has_feature('nonstripped') is False:
        return

    create_file(SOURCE, 'source.c')
    create_binary('source.c', stripped=False)
    head, *data, footer = run_rmlint('-T "none +nonstripped"')
    assert footer['total_files'] == 2
    assert footer['total_lint_size'] == 0  # We cannot determine exact lint size.
    assert data[0]['type'] == 'nonstripped'


# regression test for GitHub issue #555
def test_executable_fifo(usual_setup_usual_teardown):
    if has_feature('nonstripped') is False:
        pytest.skip("needs 'nonstripped' feature")

    fifo_path = os.path.join(TESTDIR_NAME, 'fifo')
    os.mkfifo(fifo_path)
    os.chmod(fifo_path, 0o755)

    # executable FIFO should not hang rmlint
    head, *data, footer = run_rmlint('-T nonstripped', timeout=5)
    assert footer['total_files'] == 0
    assert footer['total_lint_size'] == 0
    assert not data



#!/usr/bin/env python3
from tests.utils import *

def test_equal_files(usual_setup_usual_teardown, needs_reflink_fs):
    path_a = create_file('1234', 'a')
    path_b = create_file('1234', 'b')

    with assert_exit_code(0):
        run_rmlint(
            '--dedupe',
            path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity="")

    with assert_exit_code(0):
        run_rmlint(
            '--dedupe',
            path_a, '//', path_b,
            use_default_dir=False,
            with_json=False)


def test_different_files(usual_setup_usual_teardown, needs_reflink_fs):
    path_a = create_file('1234', 'a')
    path_b = create_file('4321', 'b')

    with assert_exit_code(1):
        run_rmlint(
            '--dedupe',
            path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity="")


def test_bad_arguments(usual_setup_usual_teardown, needs_reflink_fs):
    path_a = create_file('1234', 'a')
    path_b = create_file('1234', 'b')
    path_c = create_file('1234', 'c')
    for paths in [
            path_a,
            ' '.join((path_a, path_b, path_c)),
            ' '.join((path_a, path_a + ".nonexistent"))
    ]:
        with assert_exit_code(1):
            run_rmlint(
                '--dedupe',
                paths,
                use_default_dir=False,
                with_json=False,
                verbosity="")


def test_directories(usual_setup_usual_teardown, needs_reflink_fs):
    path_a = os.path.dirname(create_dirs('dir_a'))
    path_b = os.path.dirname(create_dirs('dir_b'))

    with assert_exit_code(1):
        run_rmlint(
            '--dedupe',
            path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity="")


def test_dedupe_works(usual_setup_usual_teardown, needs_reflink_fs):

    # test files need to be larger than btrfs node size to prevent inline extents
    path_a = create_file('1' * 100000, 'a')
    path_b = create_file('1' * 100000, 'b')

    # confirm that files are not reflinks
    with assert_exit_code(1):
        run_rmlint(
            '--is-reflink', path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity=""
        )

    # reflink our files
    with assert_exit_code(0):
        run_rmlint(
            '--dedupe', path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity=""
        )

    # confirm that they are now reflinks
    with assert_exit_code(0):
        run_rmlint(
            '--is-reflink', path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity=""
        )


def test_clone_handler(usual_setup_usual_teardown, needs_reflink_fs):
    # test files need to be larger than btrfs node size to prevent inline extents
    path_a = create_file('1' * 100000, 'a')
    path_b = create_file('1' * 100000, 'b')

    sh_path = os.path.join(TESTDIR_NAME, 'rmlint.sh')

    # generate rmlint.sh and check that it correctly selects files for cloning
    with assert_exit_code(0):
        run_rmlint(
            '-S a -o sh:{p} -c sh:clone'.format(p=sh_path),
            path_a, path_b,
            use_default_dir=False,
            with_json=False
        )

    # parse output file for expected clone command
    counts = pattern_count(sh_path, ["^clone *'", "^skip_reflink *'"])
    assert counts[0] == 1
    assert counts[1] == 0

    # now reflink the two files and check again
    with assert_exit_code(0):
        run_rmlint(
            '--dedupe', path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity=""
        )
    with assert_exit_code(0):
        run_rmlint(
            '-S a -o sh:{p} -c sh:clone'.format(p=sh_path),
            path_a, path_b,
            use_default_dir=False,
            with_json=False
        )

    counts = pattern_count(sh_path, ["^clone *'", "^skip_reflink *'"])
    assert counts[0] == 0
    assert counts[1] == 1



import os
import subprocess
from tests.utils import *


def check_is_reflink_status(status_code, *paths):
    with assert_exit_code(status_code):
        run_rmlint_once(
            '--is-reflink', *paths,
            use_default_dir=False,
            with_json=False,
            verbosity=''
        )


def test_bad_arguments(usual_setup_usual_teardown):
    path_a = create_file('xxx', 'a')
    path_b = create_file('xxx', 'b')
    path_c = create_file('xxx', 'c')
    for paths in [
        (path_a,),
        (path_a, path_b, path_c),
        (path_a, path_a + '.nonexistent')
    ]:
        check_is_reflink_status(1, *paths)  # RM_LINK_NONE


def test_directories(usual_setup_usual_teardown):
    path_a = create_dirs('dir_a')
    path_b = create_dirs('dir_b')
    check_is_reflink_status(3, path_a, path_b)  # RM_LINK_NOT_FILE


def test_different_sizes(usual_setup_usual_teardown):
    path_a = create_file('xxx', 'a')
    path_b = create_file('xxxx', 'b')
    check_is_reflink_status(4, path_a, path_b)  # RM_LINK_WRONG_SIZE


def test_same_path(usual_setup_usual_teardown):
    path_a = create_file('xxx', 'a')
    check_is_reflink_status(6, path_a, path_a)  # RM_LINK_SAME_FILE


def test_path_double(usual_setup_usual_teardown):
    path_a = create_file('xxx', 'dir/a')
    create_link('dir', 'dir_symlink', symlink=True)
    path_b = os.path.join(TESTDIR_NAME, 'dir_symlink/a')
    try:
        run_rmlint_once(
            '--is-reflink', path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity='',
        )
    except subprocess.CalledProcessError as exc:
        # XXX: this will consistently be 7 (RM_LINK_PATH_DOUBLE) in the future
        assert exc.returncode in (6, 7)  # usually 6 (RM_LINK_SAME_FILE)
    else:
        assert False


def test_hardlinks(usual_setup_usual_teardown):
    path_a = create_file('xxx', 'a')
    path_b = path_a + '_hardlink'
    create_link('a', 'a_hardlink', symlink=False)
    check_is_reflink_status(8, path_a, path_b)  # RM_LINK_HARDLINK


def test_symlink(usual_setup_usual_teardown):
    path_a = create_file('xxx', 'a')
    path_b = create_file('xxx', 'b') + '_symlink'
    create_link('b', 'b_symlink', symlink=True)
    try:
        check_is_reflink_status(9, path_a, path_b)  # RM_LINK_SYMLINK
    except AssertionError:
        pass  # expected failure
    else:
        raise AssertionError('test was epxected to fail')


def _run_dd_urandom(outfile, blocksize, count, extra=''):
    fmt = 'dd status=none oflag=sync bs={bs} count={c} {e} if=/dev/urandom'
    subprocess.run(
        [*fmt.format(bs=blocksize, c=count, e=extra).split(), 'of=' + outfile],
        check=True,
    )


def _make_reflink_testcase(extents, hole_extents=None, break_link=False):
    path_a = os.path.join(TESTDIR_NAME, 'a')
    path_b = os.path.join(TESTDIR_NAME, 'b')

    _run_dd_urandom(path_a, '4K', extents)

    if hole_extents is not None:
        os.truncate(path_a, (extents + hole_extents) * 4 * 1024)

    subprocess.run(['cp', '--reflink', path_a, path_b], check=True)

    if break_link:
        _run_dd_urandom(path_b, '4K', 1, 'seek=1 conv=notrunc')

    # expect RM_LINK_NONE or RM_LINK_REFLINK
    check_is_reflink_status(1 if break_link else 0, path_a, path_b)


# GitHub issue #527: Make sure rmlint does not skip every other extent.
def test_second_extent_differs(usual_setup_usual_teardown, needs_reflink_fs):
    _make_reflink_testcase(extents=5, break_link=True)


# GitHub issue #528, part 1: Make sure the last extent is not ignored.
def test_last_extent_differs(usual_setup_usual_teardown, needs_reflink_fs):
    _make_reflink_testcase(extents=2, break_link=True)


# GitHub issue #528, part 2: Make sure files that end in a hole can be identified as reflinked.
def test_reflink_ends_with_hole(usual_setup_usual_teardown, needs_reflink_fs):
    _make_reflink_testcase(extents=1, hole_extents=1)


def _copy_file_range(src, dst, count, offset_src, offset_dst):
    bytes_copied = os.copy_file_range(src, dst, count, offset_src, offset_dst)
    if bytes_copied < count:
        raise RuntimeError('copy_file_range only copied {} bytes (expected {})'.format(bytes_copied, count))


def kb(n):
    return 1024 * n


def _hole_testcase_inner(extents):
    path_a = os.path.join(TESTDIR_NAME, 'a')
    path_b = os.path.join(TESTDIR_NAME, 'b')

    _run_dd_urandom(path_a, kb(16) // extents, extents)
    with open(path_b, 'wb') as f:
        os.truncate(f.fileno(), kb(16))  # same-sized file with no extents

    with open(path_a, 'rb') as fsrc, open(path_b, 'wb') as fdst:
        infd, outfd = fsrc.fileno(), fdst.fileno()
        yield infd, outfd

    # expect RM_LINK_NONE
    check_is_reflink_status(1, path_a, path_b)


# GitHub issue #611: Make sure holes can be detected when the physical offsets and logical
# extent ends are otherwise the same.
def test_hole_before_extent(usual_setup_usual_teardown, needs_reflink_fs):
    for infd, outfd in _hole_testcase_inner(extents=2):
        # copy first half of first extent with 4K offset
        _copy_file_range(infd, outfd, kb(4), kb(0), kb(4))
        # copy second extent
        _copy_file_range(infd, outfd, kb(8), kb(8), kb(8))


# GitHub issue #530: Make sure physically adjacent extents aren't merged if there is a
# hole between them logically.
def test_hole_between_extents(usual_setup_usual_teardown, needs_reflink_fs):
    for infd, outfd in _hole_testcase_inner(extents=1):
        # copy first extent
        _copy_file_range(infd, outfd, kb(8), kb(0), kb(0))
        # copy first half of second extent with 4K offset
        _copy_file_range(infd, outfd, kb(4), kb(8), kb(12))


def test_default_outputs_disabled(usual_setup_usual_teardown):
    create_file('xxx', 'a')
    create_file('xxx', 'b')

    cwd = os.getcwd()
    try:
        os.chdir(TESTDIR_NAME)
        try:
            run_rmlint('--is-reflink a b', use_default_dir=False, with_json=False)
        except subprocess.CalledProcessError:
            pass  # nonzero exit status is expected

        # --is-reflink should not create or overwrite rmlint.sh or rmlint.json
        assert not os.path.exists('rmlint.sh')
        assert not os.path.exists('rmlint.json')
    finally:
        os.chdir(cwd)



#!/usr/bin/env python3
# encoding: utf-8
from tests.utils import *


# Regression test for directly passing broken symbolic links
# to the command line. See https://github.com/sahib/rmlint/pull/444
def test_bad_symlinks_as_direct_args(usual_setup_usual_teardown):
    create_file('xxx', 'a')
    create_file('xxx', 'b')

    # Create symbolic links:
    create_link('a', 'link_a', symlink=True)
    create_link('b', 'link_b', symlink=True)

    link_a_path = os.path.join(TESTDIR_NAME, 'link_a')
    link_b_path = os.path.join(TESTDIR_NAME, 'link_b')

    # Remove original files:
    os.remove(os.path.join(TESTDIR_NAME, 'a'))
    os.remove(os.path.join(TESTDIR_NAME, 'b'))

    # Directly point rmlint to symlinks, should result
    # in directly finding them.
    head, *data, footer = run_rmlint(link_a_path, link_b_path)
    assert len(data) == 2
    assert data[0]['type'] == 'badlink'
    assert data[1]['type'] == 'badlink'

    assert {data[0]['path'], data[1]['path']} == \
            {link_a_path, link_b_path}



#!/usr/bin/env python3
import os
import subprocess
import pytest

from tests.utils import *


def create_files():
    # Same size, different content.
    create_file('a', '1.a')
    create_file('b', '1.b')

    # Same size, same content.
    create_file('a', '2.a')
    create_file('a', '2.a_')

    # Different content and size
    create_file('a' * 3, '3.a')
    create_file('a' * 4, '3.a_')

    # Size group that will generate ext_cksums for all files
    create_file('b' * 2, '4.a')
    create_file('b' * 2, '4.b')
    create_file('c' * 2, '4.c')
    create_file('c' * 2, '4.d')

    # duplicate_dirs + with --write_unfinished
    create_file('x', 'dir_a/1')
    create_file('x', 'dir_b/1')


def check(data, write_cache):
    unfinished = [p['path'] for p in data if p['type'] == 'unique_file']
    dupe_files = [p['path'] for p in data if p['type'] == 'duplicate_file']
    dupe_trees = [p['path'] for p in data if p['type'] == 'duplicate_dir']

    path_in = lambda name, paths: os.path.join(TESTDIR_NAME, name) in paths

    if write_cache:
        assert len(unfinished) == 3
        assert path_in('1.b', unfinished)
        assert path_in('dir_a/1', unfinished)
        assert path_in('dir_b/1', unfinished)

    assert len(dupe_trees) == 2
    assert path_in('dir_a', dupe_trees)
    assert path_in('dir_b', dupe_trees)

    assert len(dupe_files) == 7
    assert path_in('2.a', dupe_files)
    assert path_in('2.a_', dupe_files)
    assert path_in('1.a', dupe_files)
    assert path_in('4.a', dupe_files)
    assert path_in('4.b', dupe_files)
    assert path_in('4.c', dupe_files)
    assert path_in('4.d', dupe_files)


def test_xattr_basic(usual_setup_usual_teardown):
    create_files()

    for _ in range(2):
        for write_cache in True, False:
            if write_cache:
                head, *data, footer = run_rmlint('-U -D -S pa --xattr-write')
            else:
                head, *data, footer = run_rmlint('-D -S pa --xattr-read')

            check(data, write_cache)

        head, *data, footer = run_rmlint('-D -S pa --xattr-clear')


@pytest.mark.parametrize("extra_opts", ["", "-D"])
def test_xattr_detail(usual_setup_usual_teardown, extra_opts):
    if not runs_as_root():
        # This tests need a ext4 fs which is created during the test.
        # The mount step sadly needs root privileges.
        return

    with create_special_fs("this-is-not-tmpfs") as ext4_path:
        # Keep the checksum fixed, if we change the default we don't want to
        # break this test (although I'm sure some tests will break)
        base_options = extra_opts + " -T df -S pa -a blake2b "

        path_1 = os.path.join(ext4_path, "1")
        path_2 = os.path.join(ext4_path, "2")
        path_3 = os.path.join(ext4_path, "3")
        path_4 = os.path.join(ext4_path, "4")

        create_file("abc", path_1)
        create_file("abc", path_2)
        create_file("def", path_3)
        create_file("longer", path_4)

        head, *data, footer = run_rmlint(base_options + ' --xattr-write')
        assert len(data) == 2

        xattr_1 = must_read_xattr(path_1)
        xattr_2 = must_read_xattr(path_2)
        xattr_3 = must_read_xattr(path_3)
        assert xattr_1["user.rmlint.blake2b.cksum"] == \
                b'ba80a53f981c4d0d6a2797b69f12f6e94c212f14685ac4b74b12bb6fdbffa2d17d87c5392aab792dc252d5de4533cc9518d38aa8dbf1925ab92386edd4009923\x00'
        assert xattr_1 == xattr_2

        # no --write-unfinished given.
        assert xattr_3 == {}

        # Repeating the caching option should have no effect on the output.
        for _ in range(10):
            head, *data, footer = run_rmlint(base_options + ' --xattr')
            # one more due to the unique_file
            assert len(data) == 3

            xattr_1 = must_read_xattr(path_1)
            xattr_2 = must_read_xattr(path_2)
            xattr_3 = must_read_xattr(path_3)
            assert xattr_1["user.rmlint.blake2b.cksum"] == \
                    b'ba80a53f981c4d0d6a2797b69f12f6e94c212f14685ac4b74b12bb6fdbffa2d17d87c5392aab792dc252d5de4533cc9518d38aa8dbf1925ab92386edd4009923\x00'
            assert xattr_1 == xattr_2

            # --write-unfinished will also write the unfinished one.
            xattr_3 = must_read_xattr(path_3)
            assert xattr_3["user.rmlint.blake2b.cksum"] == \
                    b'36badf2227521b798b78d1bd43c62520a35b9b541547ff223f35f74b1168da2cd3c8d102aaee1a0cc217b601258d80151067cdee3a6352517b8fc7f7106902d3\x00'

            # unique file which was not hashed -> does not need to be touched.
            xattr_4 = must_read_xattr(path_4)
            assert xattr_4 == {}

        # Try clearing the attributes:
        head, *data, footer = run_rmlint(base_options + '--xattr-clear')
        assert len(data) == 2
        assert must_read_xattr(path_1) == {}
        assert must_read_xattr(path_2) == {}
        assert must_read_xattr(path_3) == {}
        assert must_read_xattr(path_4) == {}


# regression test for GitHub issue #475
# NB: this test is only effective if RM_TS_DIR is on an xattr-capable filesystem
def test_treemerge_xattr_hardlink(usual_setup_usual_teardown):
    create_file('xxx', 'a/x')
    create_file('yyy', 'a/y')
    create_file('xxx', 'b/x')
    create_file('yyy', 'b/y')

    sh_path = os.path.join(TESTDIR_NAME, 'rmlint.sh')
    head, *data, foot = run_rmlint('--xattr-write -o sh:{p} -c sh:hardlink'.format(p=sh_path))
    assert len(data) == 4

    # run script to hardlink files
    subprocess.check_output([sh_path, '-d'])

    # This used to fail with 'rm_shred_group_free: assertion failed: (self->num_pending == 0)'
    head, *data, foot = run_rmlint('-D --xattr-read')
    assert len(data) == 6


# NB: this test is only effective if RM_TS_DIR is on an xattr-capable filesystem
@pytest.mark.parametrize("clamp", ['-q 1', '-Q 1', '-q 50%', '-Q 50%'])
def test_clamp_xattr_false_negative(usual_setup_usual_teardown, clamp):
    create_file('xxx', 'a')
    create_file('yyy', 'b')

    # we used to write xattrs even when clamping is used
    head, *data, foot = run_rmlint('--xattr', clamp)
    assert all(e['type'] == 'unique_file' for e in data)

    create_file('xxx', 'c')

    # the first run after creating 'c' is ok...
    head, *data, foot = run_rmlint('--xattr', force_no_pendantic=True)
    assert len([e for e in data if e['type'] == 'duplicate_file']) == 2  # 'a' matches 'c'

    # but we would get a false negative here, as the xattrs didn't match
    head, *data, foot = run_rmlint('--xattr', force_no_pendantic=True)
    assert len([e for e in data if e['type'] == 'duplicate_file']) == 2  # do they still match?


# NB: this test is only effective if RM_TS_DIR is on an xattr-capable filesystem
@pytest.mark.parametrize("clamp", ['-q 2', '-Q 1', '-q 70%', '-Q 50%'])
def test_clamp_xattr_false_positive(usual_setup_usual_teardown, clamp):
    # directories 'a' and 'b' obviously do not match
    # extra files are needed to satisfy preprocessing, which compares file size
    create_file('xxx', '1')
    create_file('xxx', 'a/1')
    create_file('x', '2')
    create_file('x', 'b/2')

    # we used to write xattrs even when clamping is used
    head, *data, foot = run_rmlint('--xattr --size 3', clamp)
    assert len([e for e in data if e['type'] == 'duplicate_file']) == 2  # '1' matches 'a/1'

    # fill in other xattrs
    head, *data, foot = run_rmlint('--xattr', force_no_pendantic=True)
    assert len([e for e in data if e['type'] == 'duplicate_file']) == 4  # '1' matches 'a/1', '2' matches 'b/2'

    # we would get a false positive here, as the xattrs matched
    head, *data, foot = run_rmlint('--xattr -T dd', force_no_pendantic=True)
    assert not any(e['type'] == 'duplicate_dir' for e in data)  # do 'a' and 'b' match?



#!/usr/bin/env python3
import os

from tests.utils import *

def test_equal_files(usual_setup_usual_teardown):
    path_a = create_file('1234', 'a')
    path_b = create_file('1234', 'b')

    with assert_exit_code(0):
        head, *data, footer = run_rmlint(
            '--equal', path_a, path_b,
            use_default_dir=False
        )

    with assert_exit_code(0):
        head, *data, footer = run_rmlint(
            '-p', '--equal', path_a, path_b,
            use_default_dir=False
        )

    # If --equal finds that both files are not equal,
    # it would return 1 as exit code which would cause
    # a CalledProcessError in run_rmlint
    assert data[0]['path'].endswith('a')
    assert data[1]['path'].endswith('b')

    path_c = create_file('1234', 'c')
    with assert_exit_code(0):
        head, *data, footer = run_rmlint(
            '--equal', path_a, path_b, path_c,
            use_default_dir=False
        )

    path_d = create_file('diff', 'd')
    with assert_exit_code(1):
        head, *data, footer = run_rmlint(
            '--equal', path_a, path_b, path_d, path_c,
            use_default_dir=False
        )


def test_no_arguments(usual_setup_usual_teardown):
    with assert_exit_code(1):
        head, *data, footer = run_rmlint(
            '--equal',
            use_default_dir=False
        )


def test_one_arguments(usual_setup_usual_teardown):
    path = create_file('1234', 'a')
    with assert_exit_code(1):
        head, *data, footer = run_rmlint(
            '--equal', path,
            use_default_dir=False
        )

    with assert_exit_code(1):
        head, *data, footer = run_rmlint(
            '--equal', path, "//",
            use_default_dir=False
        )

    with assert_exit_code(1):
        head, *data, footer = run_rmlint(
            '--equal', "//", path,
            use_default_dir=False
        )


def test_equal_directories(usual_setup_usual_teardown):
    path_a = os.path.dirname(create_file('xxx', 'dir_a/x'))
    path_b = os.path.dirname(create_file('xxx', 'dir_b/x'))

    with assert_exit_code(0):
        head, *data, footer = run_rmlint(
            '--equal', path_a, path_b,
            use_default_dir=False
        )

    with assert_exit_code(0):
        head, *data, footer = run_rmlint(
            '-p', '--equal', path_a, path_b,
            use_default_dir=False
        )


def test_dir_and_file(usual_setup_usual_teardown):
    path_a = os.path.dirname(create_file('xxx', 'dir_a/x'))
    path_b = create_file('xxx', 'x')

    # This should fail since we should not mix directories with files,
    # even if they have the same content.
    with assert_exit_code(1):
        head, *data, footer = run_rmlint(
            '--equal', path_a, path_b,
            use_default_dir=False,
        )


# Regression test for Issue #233
def test_equal_hidden_dirs(usual_setup_usual_teardown):
    path_a = os.path.dirname(create_file('xxx', 'dir_a/x'))
    path_b = os.path.dirname(create_file('xxx', '.dir_b/.x'))

    # This should fail since we should not mix directories with files,
    # even if they have the same content.
    with assert_exit_code(0):
        head, *data, footer = run_rmlint(
            '--equal', path_a, path_b,
            use_default_dir=False,
        )


# Regression test for Issue #234
def test_equal_empty_files_or_other_lint(usual_setup_usual_teardown):
    path_a = create_file('', 'x')
    path_b = create_file('', 'y')

    # This should fail since we should not mix directories with files,
    # even if they have the same content.
    with assert_exit_code(0):
        head, *data, footer = run_rmlint(
            '--equal', path_a, path_b,
            use_default_dir=False,
        )


# regression test for GitHub issue #552
def test_default_outputs_disabled(usual_setup_usual_teardown):
    create_file('xxx', 'a')
    create_file('xxx', 'b')

    cwd = os.getcwd()
    try:
        os.chdir(TESTDIR_NAME)
        run_rmlint('--equal a b', use_default_dir=False, with_json=False)

        # Users of --equal, including our own sh format, do not expect to
        # create or overwrite rmlint.sh or rmlint.json.
        assert not os.path.exists('rmlint.sh')
        assert not os.path.exists('rmlint.json')
    finally:
        os.chdir(cwd)



#!/usr/bin/env python3

from setuptools import setup
from setuptools.command.install import install
from distutils.command.install_data import install_data

import os
import logging
import subprocess

GRESOURCE_DIR = 'shredder/resources'
GRESOURCE_FILE = 'shredder.gresource.xml'
GSCHEMA_DIR_SUFFIX = 'share/glib-2.0/schemas'


def read_version():
    vfp = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, '.version')
    with open(vfp, 'r') as handle:
        version_string = handle.read()

    version_numbers, _ = version_string.split(' ', 1)
    return version_numbers


class install_glib_resources(install):
    def run(self):
        self._build_gresources()
        super().run()

    def _build_gresources(self):
        '''
        Compile the resource bundle
        '''
        logging.info('==> Calling glib-compile-resources')
        try:
            subprocess.call([
                'glib-compile-resources',
                '--sourcedir={}'.format(GRESOURCE_DIR),
                os.path.join(GRESOURCE_DIR, GRESOURCE_FILE)
            ])
        except subprocess.CalledProcessError as err:
            logging.error('==> Failed :(')


class compile_glib_schemas(install_data):
    def run(self):
        super().run()
        if os.environ.get("COMPILE_GLIB_SCHEMA", False):
            self._build_gschemas()
        else:
            self.print_compile_instructions()

    def gschema_dir(self):
        return os.path.join(self.install_dir, GSCHEMA_DIR_SUFFIX)

    def print_compile_instructions(self):
        logging.info('==> You may need to compile glib schemas manually:\n')
        logging.info('    sudo glib-compile-schemas {}\n'.format(
            self.gschema_dir()))

    def _build_gschemas(self):
        '''
        Make sure the schema file is updated after installation,
        otherwise the gui will trace trap.
        '''
        logging.info('==> Compiling GLib Schema files')
        try:
            subprocess.call(['glib-compile-schemas', self.gschema_dir()])
            logging.info('==> OK!')
        except subprocess.CalledProcessError as err:
            logging.error('==> Could not update schemas: ', err)
            self.print_compile_instructions()


setup(
    name='Shredder',
    version=read_version(),
    description='A gui frontend to rmlint',
    long_description='A graphical user interface to rmlint using GTK+',
    author='Christopher Pahl',
    author_email='sahib@online.de',
    url='https://rmlint.rtfd.org',
    platforms='any',
    cmdclass={
        'install': install_glib_resources,
        'install_data': compile_glib_schemas
    },
    packages=['shredder', 'shredder.views'],
    package_data={'': [
        'resources/*.gresource'
    ]},
    data_files=[
        (
            'share/icons/hicolor/scalable/apps',
            ['shredder/resources/shredder.svg']
        ),(
            'share/glib-2.0/schemas',
            ['shredder/resources/org.gnome.Shredder.gschema.xml']
        ),(
            'share/applications',
            ['shredder.desktop']
        ),
    ],
    scripts=['bin/shredder'],
)



#!/usr/bin/env python3
# encoding: utf-8
from tests.utils import *

# These tests are only here to check if printing help works.
# Well, actually it's to increase coverage to be honest.


def test_help(usual_setup_usual_teardown):
    yelp = subprocess.check_output(
        ['./rmlint', '--help'], stderr=subprocess.STDOUT
    ).decode('utf-8')
    assert 'man 1 rmlint' in yelp
    assert '--show-man' in yelp


def test_man(usual_setup_usual_teardown):
    yelp = subprocess.check_output(
        ['./rmlint', '--show-man'], stderr=subprocess.STDOUT
    ).decode('utf-8')
    assert 'Pahl' in yelp
    assert 'Thomas' in yelp



from tests.utils import *


def test_color(usual_setup_usual_teardown):
    # color disabled by pipe
    output = run_rmlint('-o summary', with_json=False, directly_return_output=True)
    assert b'\x1b[' not in output

    # color cannot be enabled with pipe
    output = run_rmlint('-o summary --with-color', with_json=False, directly_return_output=True)
    assert b'\x1b[' not in output


def test_no_color_crash(usual_setup_usual_teardown):
    # cause rmlint to fail with a simple error message, without color
    result, _ = run_rmlint_once('-kK --no-with-color', with_json=False, check=False, verbosity='')
    assert result.returncode == 1  # should not be a segfault or any other fatal signal
    assert result.stderr
    assert b'\x1b[' not in result.stderr
